unit TMinMaxClass;

interface

uses Classes, TSmallRamugenClasses, SysUtils, StdCtrls, ExtCtrls, Spin;

type

TMinMax = class(TPersistent)
          protected
           FHardMin,
            FHardMax,
            FMin,
            FMax,
            FChosen,
            FUp,
            FDown : Integer;
           FDirection : TGlobalProbability; // =probability of moving UP instead of DOWN
           AChange : TGlobalProbability;
           procedure SetMin(T : Integer);
           procedure SetMax(T : Integer);
           procedure SetHardMax(T : Integer);
           procedure SetHardMin(T : Integer);
           function GetProb : Boolean;
          public
           constructor Create;
           destructor Destroy;  override;
           procedure Assign(aMinMax : TPersistent); override;
           procedure AssignTo(aMinMax : TPersistent); override;
           procedure SetRange (KMin, KMax : Integer); overload; //most flexible setter
           procedure SetRange (SMin, SMax : String); overload;
           procedure SetRange (EMin, EMax : TObject); overload;
           function Choose (XMin : Integer = -1; XMax : Integer = -1) : Integer; virtual;
           property HardMax : Integer read FHardMax write SetHardMax;
           property HardMin : Integer read FHardMin write SetHardMin;
          published
           property MoveUp : Integer read FUp write FUp;
           property MoveDown : Integer read FDown write FDown;
           property Min : Integer read FMin write SetMin;
           property Max : Integer read FMax write SetMax;
           property LastChosen : Integer read FChosen;
           property MovingUp : Boolean read GetProb;
           property Change : TGlobalProbability read AChange write AChange;
          end;


TSimpleMinMax = class(TMinMax)
           protected
            FUnitSize : Integer;
            procedure SetUnitSize(T : Integer);
           public
            constructor Create;
            procedure Assign(aMinMax : TPersistent); override;
            procedure AssignTo(aMinMax : TPersistent); override;
            function Choose (XMin : Integer = -1; XMax : Integer = -1) : Integer; override;
           published
            property UnitSize : Integer read FUnitSize write SetUnitSize;
          end;

TExtendedMinMax = class(TMinMax)
                  protected
                   FWeighted : Boolean;
                   AChosen : Integer;
                   procedure SetTitle(T : string);
                   function GetTitle : string;
                   procedure SetDesc(T : string);
                   function GetDesc : string;
                  public
                   Weightings : TWeightingSystem;
                   constructor Create; 
                   destructor Destroy;   override;
                   procedure Assign(aMinMax : TPersistent); overload; override;
                    procedure Assign(Obj : TObject); reintroduce; overload;
                    procedure Assign(A : Array of Integer); reintroduce; overload;
                   procedure AssignTo(aMinMax : TPersistent); overload; override;
                    procedure AssignTo(Obj : TObject); reintroduce; overload;
                    procedure AssignTo(A : Array of Integer); reintroduce; overload;

                   function Choose (XMin : Integer = -1; XMax : Integer = -1) : Integer; override;
                     //TWeightingSystem method adapters
                    procedure Mu;
                    procedure Initialize (X : Integer);
                    procedure InputWeightings(const ShowRulers : Boolean = False);
                    property LastChosen : Integer read AChosen;
                  published
                   property Weighted : Boolean read FWeighted write FWeighted;
                    //TWeightingSystem property adapters
                   property Title : string read GetTitle write SetTitle;
                   property Description : string read GetDesc write SetDesc;
                  end;

implementation

//constructors
constructor TMinMax.Create;
begin
 inherited;
 FDirection := TGlobalProbability.Create;
 AChange := TGlobalProbability.Create;
 AChange.Probability := 1;
 FHardMin := 0;
 FHardMax := 127;
 FMin := 0; FMax := 127;
 FUp := 50;
 FDown := 50;
 randomize;
end;

destructor TMinMax.Destroy;
begin
 FreeAndNil(FDirection);
 FreeAndNil(AChange);
 inherited;
end;

constructor TSimpleMinMax.Create;
begin
 inherited;
 FUnitSize := 1;
end;

constructor TExtendedMinMax.Create;
begin
 inherited;
 Weightings := TWeightingSystem.Create(nil);
 SetDesc('Use the boxes below to enter weighted bias-amounts to assist the selection of values');
end;

destructor TExtendedMinMax.Destroy;
begin
 FreeAndNil(Weightings);
 inherited;
end;

//Mu

procedure TExtendedMinMax.Mu;
begin
 FWeighted := true;
 Weightings.Mu;
end;

//Assign

procedure TMinMax.Assign(aMinMax : TPersistent);
begin
 if aMinMax is TMinMax then begin
                             HardMax := TMinMax(aMinMax).HardMax;
                             HardMin := TMinMax(aMinMax).HardMin;
                             MoveUp := TMinMax(aMinMax).MoveUp;
                             MoveDown := TMinMax(aMinMax).MoveDown;
                             Min := TMinMax(aMinMax).Min;
                             Max := TMinMax(aMinMax).Max;
                             Change.Probability := TMinMax(aMinMax).Change.Probability;
                               //nb. don't need to do "direction", as it's a calculated field
                            end
                       else
                            inherited Assign(aMinMax);
end;

procedure TMinMax.AssignTo(aMinMax : TPersistent);
begin
 if aMinMax is TMinMax then begin
                             TMinMax(aMinMax).HardMax := HardMax;
                             TMinMax(aMinMax).HardMin := HardMin;
                             TMinMax(aMinMax).MoveUp := MoveUp;
                             TMinMax(aMinMax).MoveDown := MoveDown;
                             TMinMax(aMinMax).Min := Min;
                             TMinMax(aMinMax).Max := Max;
                             TMinMax(aMinMax).Change.Probability := Change.Probability;
                               //nb. don't need to do "direction", as it's a calculated field
                            end
                       else
                            inherited AssignTo(aMinMax);
end;


procedure TExtendedMinMax.Assign(aMinMax : TPersistent);
begin
 if aMinMax is TExtendedMinMax then begin
                                   Weighted := TExtendedMinMax(aMinMax).Weighted;
                                   Weightings.Assign(TExtendedMinMax(aMinMax).Weightings);
                                  end;
 inherited Assign(aMinMax);
end;

//Overloaded Assign methods which are actually delegation adapters for
//TWeightingSystem worker object

procedure TExtendedMinMax.Assign(Obj : TObject);
begin
 Weightings.Assign(Obj);
end;

procedure TExtendedMinMax.Assign(A : array of Integer);
begin
 Weightings.Assign(A);
end;


procedure TExtendedMinMax.AssignTo(aMinMax : TPersistent);
begin
 if aMinMax is TExtendedMinMax then begin
                                   TExtendedMinMax(aMinMax).Weighted := Weighted;
                                   Weightings.AssignTo(TExtendedMinMax(aMinMax).Weightings);
                                  end;
 inherited AssignTo(aMinMax);
end;

procedure TExtendedMinMax.AssignTo(Obj : TObject);
begin
 Weightings.AssignTo(Obj);
end;

procedure TExtendedMinMax.AssignTo(A : array of Integer);
begin
 Weightings.AssignTo(A);
end;


procedure TSimpleMinMax.Assign(aMinMax : TPersistent);
begin
 if aMinMax is TSimpleMinMax then begin
                                   UnitSize := TSimpleMinMax(aMinMax).UnitSize;
                                  end;
 inherited Assign(aMinMax);
end;

procedure TSimpleMinMax.AssignTo(aMinMax : TPersistent);
begin
 if aMinMax is TSimpleMinMax then begin
                                   TSimpleMinMax(aMinMax).UnitSize := UnitSize;
                                  end;
 inherited AssignTo(aMinMax);
end;

//Setters


procedure TMinMax.SetMin(T : Integer);
begin
 if T < FHardMin then T := FHardMin; if T > FHardMax then T := FHardMax;
 FMin := T;
 if FMin > FMax then FMax := FMin;
end;

procedure TMinMax.SetMax(T : Integer);
begin
 if T < FHardMin then T := FHardMin; if T > FHardMax then T := FHardMax;
 FMax := T;
 if FMax < FMin then FMin := FMax;
end;

//not strictly speaking a setter, hence the overload
procedure TMinMax.SetRange (KMin, KMax : Integer);
var x : Integer;
begin
 if KMin < FHardMin then KMin := FHardMin; if KMax > FHardMax then KMax := FHardMax;
 if KMin > KMax then begin
                      x := KMin;
                      KMin := KMax;
                      KMax := x;
                     end;
 FMin := KMin;
 FMax := KMax;
end;

procedure TMinMax.SetRange (SMin, SMax : String);
var x, KMin, KMax : Integer;
begin
 Val(SMin, KMin, x); if x <> 0 then exit;
 Val(SMax, KMax, x); if x <> 0 then exit;
 SetRange(KMin, KMax);
end;

procedure TMinMax.SetRange (EMin, EMax : TObject);
var KMin, KMax : Integer;
     function ConvertObject(Obj : TObject) : Integer; //errorcode = -1
     var Y,Z : Integer;
     begin
       Y := -1;
       if Obj is TEdit then Val(TEdit(Obj).Text, Y, Z);
       if Obj is TLabel then Val(TLabel(Obj).Caption, Y, Z);
       if Obj is TSpinEdit then Y := TSpinEdit(Obj).Value;
       Result := Y;
     end;
begin
 KMin := ConvertObject(EMin); if KMin = -1 then exit;
 KMax := ConvertObject(EMax); if KMax = -1 then exit;
 SetRange(KMin, KMax);
end;


procedure TMinMax.SetHardMax(T : Integer);
begin
 FHardMax := T;
 if FMax > FHardMax then FMax := T;
 if FMin > FHardMax then FMin := T;
end;

procedure TMinMax.SetHardMin(T : Integer);
begin
 FHardMin := T;
 if FMax < FHardMin then FMax := T;
 if FMin < FHardMin then FMin := T;
end;


procedure TSimpleMinMax.SetUnitSize(T : Integer);
begin
 if T < 1 then T := 1;
 if T > FHardMax then T := FHardMax;
 FUnitSize := T;
end;


procedure TExtendedMinMax.SetTitle(T : String);
begin
 Weightings.Title := T;
end;

procedure TExtendedMinMax.SetDesc(T : String);
begin
 Weightings.Description := T;
end;

//Getters

function TMinMax.GetProb: Boolean;
begin
  if FDown = 0 then
                    FDirection.Probability := 0
               else
                    FDirection.Probability := FUp / (FUp + FDown);
  Result := FDirection.IsRequired;
end;

function TExtendedMinMax.GetTitle : String;
begin
 Result := Weightings.Title;
end;

function TExtendedMinMax.GetDesc : String;
begin
 Result := Weightings.Description;
end;



//Adapters

procedure TExtendedMinMax.Initialize (X : Integer);
begin
 Weightings.Initialize(X);
end;

procedure TExtendedMinMax.InputWeightings(const ShowRulers : Boolean = False);
begin
 Weightings.InputWeightings(FMin, FMax, ShowRulers);
end;


//Full-on methods

function TExtendedMinMax.Choose (XMin : Integer = -1; XMax : Integer = -1) : Integer;  //partial adapter
begin
 if XMin = -1 then XMin := FMin;
 if XMax = -1 then XMax := FMax;
 if Weighted then Result := Weightings.Choose (XMin, XMax) else
                  Result := inherited Choose (Xmin, XMax);
 FChosen := Result;
end;


function TMinMax.Choose (XMin : Integer = -1; XMax : Integer = -1) : Integer;
begin
 if XMin = -1 then XMin := FMin;
 if XMax = -1 then XMax := FMax;
 Result := trunc(random(XMax - XMin + 1)) + XMin;
end;

function TSimpleMinMax.Choose (XMin : Integer = -1; XMax : Integer = -1) : Integer;
begin
 if XMin = -1 then XMin := FMin;
 if XMax = -1 then XMax := FMax;
 FChosen := inherited Choose(XMin, XMax);
 FChosen := FChosen - (FChosen mod FUnitSize);
 Result := FChosen;
end;


end.
